/************************************************************************/
/* @\F pPbg_                                              */
/*                                                                      */
/* TvF pPbgs܂                                        */
/*        trclkɓ܂                                             */
/*                                                                      */
/*                                                                  */
/*                                                                      */
/*                                                                      */
/* ύXF                                                           */
/*  +------------------ ԍ (000 ` 999)                           */
/*  |    +------------- VKAύXAǉA폜̕                    */
/*  v    v                                                              */
/*  No     N                                               */
/* ----+------+----------+--------------------------------------------- */
/*  000 VK   10/04/26                                                 */
/*  001    12/12/26   xsynclQ[gԒZk                         */
/*                                                                      */
/************************************************************************/

module packet_gen(
	trclk,
	reset,
	dataout,
	empty,
	full_det,
	rd,
	fifo_rst,
	clk,
	xsync,
	d
	);

/***************************************************************************/
/* DECLARE THE PORT DIRECTION                                              */
/***************************************************************************/

 input		trclk;
 input		reset;
 input	[35:0]	dataout;
 input		empty;
 input		full_det;
 output		rd;
 output		fifo_rst;
 output		clk;
 output		xsync;
 output	[3:0]	d;


/***************************************************************************/
/* DECLARE THE NODE(WIRE/REG)                                              */
/***************************************************************************/
 reg		full_det_1l;
 reg		full_sync;

 reg		fifo_rd;
 reg		fifo_clr_rq;
 reg	[1:0]	clr_cnt;
 reg	[35:0]	fifo_dat;
 reg		lost;
 reg	[2:0]	fifo_st;

 reg		fifo_dat_valid;

 reg	[3:0]	o_cnt;
 reg	[35:0]	rd_dat;
 reg		adr_keep;
 reg		xvalid;
 reg	[3:0]	pkt_d;
 reg		pkt_end;
 reg	[1:0]	pkt_st;

 wire		packet_type;

/***************************************************************************/
/* patameter                                                               */
/***************************************************************************/
parameter
	WAIT_EMPTY_NEG	= 3'b000,
	WAIT_RD_DAT		= 3'b001,
	GET_FIFO_DAT	= 3'b010,
	WAIT_PKT_END	= 3'b011,
	FIFO_CLR		= 3'b100;

parameter
	WAIT_FIFO_DAT	= 2'b00,
	ADR_OUT0		= 2'b01,
	ADR_OUT1		= 2'b10,
	DAT_OUT			= 2'b11;

parameter
	bus_size	= 2'b10;

parameter
	packet_size	= 4'b1000;

parameter
	fifo_clr_time	= 2'b11;


/***************************************************************************/
/* output                                                                  */
/***************************************************************************/
	assign clk		= ~trclk ;
	assign xsync	= xvalid ;
	assign d		= pkt_d  ;
	assign rd		= fifo_rd;
	assign fifo_rst	= fifo_clr_rq | reset;


/***************************************************************************/
/* synchronize                                                             */
/***************************************************************************/
	always @(posedge trclk or posedge reset) begin
		if (reset) begin
			full_det_1l	<= 1'b0;
			full_sync	<= 1'b0;
		end
		else begin
			full_det_1l	<= full_det;
			full_sync	<= full_det_1l;
		end
	end

/***************************************************************************/
/* Packet Generator                                                        */
/***************************************************************************/

	always @(posedge trclk or posedge reset) begin
		if (reset) begin
			fifo_rd			<= 1'b0;
			fifo_clr_rq		<= 1'b0;
			clr_cnt			<= 0;
			fifo_dat[35:0]	<= 0;
			lost			<= 1'b0;
			fifo_st			<= WAIT_EMPTY_NEG;
		end
		else begin
			case (fifo_st)
			WAIT_EMPTY_NEG : begin
				fifo_clr_rq		<= 1'b0;
				clr_cnt			<= fifo_clr_time;
				lost			<= lost;
				if ( full_sync ==  1'b1 ) begin
					fifo_rd		<= 1'b0;
					fifo_st		<= FIFO_CLR;
				end
				else if ( empty == 1'b0 ) begin
					fifo_rd		<= 1'b1;
					fifo_st		<= WAIT_RD_DAT;
				end
				else begin
					fifo_rd		<= 1'b0;
					fifo_st		<= WAIT_EMPTY_NEG;
				end
			end
			WAIT_RD_DAT : begin
				fifo_rd			<= 1'b0;
				fifo_clr_rq		<= 1'b0;
				lost			<= lost;
				fifo_st			<= GET_FIFO_DAT;
			end
			GET_FIFO_DAT : begin
				fifo_rd			<= 1'b0;
				fifo_clr_rq		<= 1'b0;
				fifo_dat[35:0]	<= dataout[35:0];
				lost			<= lost;
				fifo_st			<= WAIT_PKT_END;
			end
			WAIT_PKT_END : begin
				if ( pkt_end ) begin
					fifo_rd			<= 1'b0;
					fifo_clr_rq		<= 1'b0;
					fifo_dat[35:0]	<= fifo_dat[35:0];
					lost			<= 1'b0;
					fifo_st			<= WAIT_EMPTY_NEG;
				end
				else begin
					fifo_rd			<= 1'b0;
					fifo_clr_rq		<= 1'b0;
					fifo_dat[35:0]	<= fifo_dat[35:0];
					lost			<= lost;
					fifo_st			<= WAIT_PKT_END;
				end
			end
			FIFO_CLR : begin
				if (clr_cnt == 2'b00) begin
					fifo_clr_rq	<= 1'b0;
					clr_cnt		<= 0;
					lost		<= lost;
					fifo_st		<= WAIT_EMPTY_NEG;
				end
				else begin
					fifo_clr_rq	<= 1'b1;
					clr_cnt		<= clr_cnt - 1;
					lost		<= 1'b1;
					fifo_st		<= FIFO_CLR;
				end
			end
			default : begin
				fifo_rd			<= 1'b0;
				fifo_clr_rq		<= 1'b0;
				clr_cnt			<= 0;
				lost			<= 1'b0;
				fifo_st			<= WAIT_EMPTY_NEG;
			end
			endcase
		end
	end

	always @(posedge trclk or posedge reset) begin
		if (reset) begin
			fifo_dat_valid	<= 1'b0;
		end
		else begin
			fifo_dat_valid	<= ( fifo_st == GET_FIFO_DAT );
		end
	end

	assign	packet_type	= ( fifo_dat[3:2] == 2'b00 ) ? 1'b0 :
						  ( fifo_dat[3:2] == 2'b01 ) ? 1'b0 : 1'b1;


	always @(posedge trclk or posedge reset) begin
		if (reset) begin
			o_cnt			<= 0;
			rd_dat[35:0]	<= 0;
			adr_keep		<= 0;
			xvalid			<= 1'b1;
			pkt_d[3:0]		<= 0;
			pkt_end			<= 1'b0;
			pkt_st			<= WAIT_FIFO_DAT;
		end
		else begin
			case (pkt_st)
			WAIT_FIFO_DAT : begin
				if ( fifo_dat_valid ) begin
					o_cnt			<= packet_size;
					rd_dat[35:0]	<= fifo_dat[35:0];
					adr_keep		<= fifo_dat[3];
					xvalid			<= 1'b1;
					pkt_d[3:0]		<= { lost, packet_type, bus_size };
					pkt_end			<= 1'b0;
					pkt_st			<= ADR_OUT0;
				end
				else begin
					o_cnt			<= 0;
					rd_dat[35:0]	<= rd_dat[35:0];
					adr_keep		<= adr_keep;
					xvalid			<= 1'b1;
//@001					pkt_d[3:0]		<= 0;
					pkt_d[3:0]		<= { 2'b00, bus_size };	//@001
					pkt_end			<= 1'b0;
					pkt_st			<= WAIT_FIFO_DAT;
				end
			end
			ADR_OUT0 : begin
				rd_dat[35:0]	<= { 4'b0000, rd_dat[35:4] };
				adr_keep		<= adr_keep;
				xvalid			<= 1'b0;
				pkt_d			<= { rd_dat[2:0], 1'b0 };
				pkt_end			<= 1'b0;
				if (o_cnt == 4'b0000 ) begin
					o_cnt	<= 0;
					pkt_st	<= WAIT_FIFO_DAT;
				end
				else begin
					o_cnt	<= o_cnt - 1;
					pkt_st	<= ( packet_type ) ? ADR_OUT1 : DAT_OUT;
				end
			end
			ADR_OUT1 : begin
				rd_dat[35:0]	<= rd_dat[35:0];
				adr_keep		<= adr_keep;
				xvalid			<= 1'b0;
				pkt_d			<= { 3'b000, adr_keep };
				pkt_end			<= 1'b0;
				if (o_cnt == 4'b0000 ) begin
					o_cnt	<= 0;
					pkt_st	<= WAIT_FIFO_DAT;
				end
				else begin
					o_cnt	<= o_cnt;
					pkt_st	<= DAT_OUT;
				end
			end
			DAT_OUT : begin
				rd_dat[35:0]	<= { 4'b0000, rd_dat[35:4] };
				adr_keep		<= adr_keep;
				xvalid			<= 1'b0;
				pkt_d			<= rd_dat[3:0];
				if (o_cnt == 4'b0000 ) begin
					o_cnt	<= 0;
//@001					pkt_end	<= 1'b1;
					pkt_end	<= 1'b0;	//@001
					pkt_st	<= WAIT_FIFO_DAT;
				end
				else begin
					o_cnt	<= o_cnt - 1;
//@001					pkt_end	<= 1'b0;
					pkt_st	<= DAT_OUT;
//@001 add begin
					if (o_cnt == 4'b0100 ) begin
						pkt_end	<= 1'b1;
					end
					else begin
						pkt_end	<= 1'b0;
					end
//@001 add end
				end
			end
			default : begin
				o_cnt			<= 0;
				rd_dat[35:0]	<= 0;
				adr_keep		<= 0;
				xvalid			<= 1'b1;
				pkt_d[3:0]		<= 0;
				pkt_end			<= 1'b0;
				pkt_st			<= WAIT_FIFO_DAT;
			end
			endcase
		end
	end


endmodule


