/************************************************************************/
/* @\F ͒i                                                    */
/*                                                                      */
/* TvF CgANZX̓AL̒os܂              */
/*        busclkɓ܂                                            */
/*                                                                      */
/*                                                                  */
/*                                                                      */
/*                                                                      */
/* ύXF                                                           */
/*  +------------------ ԍ (000 ` 999)                           */
/*  |    +------------- VKAύXAǉA폜̕                    */
/*  v    v                                                              */
/*  No     N                                               */
/* ----+------+----------+--------------------------------------------- */
/*  000 VK   13/03/15                                                 */
/*                                                                      */
/************************************************************************/


module sync (
	busclk,
	reset,
	xtwr,
	xtbe,
	xtcs,
	ta,
	td,
	datain,
	wr,
	empty,
	full,
	fifo_rst,
	full_det
	);

/***************************************************************************/
/* DECLARE THE PORT DIRECTION                                              */
/***************************************************************************/

 input		busclk;
 input		reset;
 input		xtwr;
 input	[3:0]	xtbe;
 input		xtcs;
 input	[4:2]	ta;
 input	[31:0]	td;
 output	[35:0]	datain;
 output		wr;
 input		empty;
 input		full;
 input		fifo_rst;
 output		full_det;


/***************************************************************************/
/* DECLARE THE NODE(WIRE/REG)                                              */
/***************************************************************************/
 reg	[1:0]	write_msk_st;

 reg 		header_det;
 reg 		wr_msk;
 reg 	[1:0]	msk_count;

 reg	[4:2]	ta_1l,ta_2l,ta_3l;
 reg	[31:0]	td_1l,td_2l,td_3l;
 reg		wr_1l,wr_2l,wr_3l;
 reg		cs_1l,cs_2l,cs_3l;
 reg		tbe_1l;
 reg	[35:0]	sync_data;
 reg		sync_wr;
 reg		full_lat;
 reg		empty_1l,empty_2l;
 reg		full_sync;

/***************************************************************************/
/* patameter                                                               */
/***************************************************************************/
parameter
 WAIT_FULL     = 2'b00,
 WAIT_FULL_NEG = 2'b01,
 WRITE_MSK     = 2'b10,
 WAIT_HEADER   = 2'b11;

/***************************************************************************/
/* output                                                                  */
/***************************************************************************/
	assign	datain = sync_data;
	assign	wr = sync_wr;
	assign	full_det = full_lat;

/***************************************************************************/
/* pipeline 1'st                                                           */
/***************************************************************************/
	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			wr_1l	<= 1'b0;
			tbe_1l	<= 1'b0;
			cs_1l	<= 1'b0;
			ta_1l	<= 3'b0;
			td_1l	<= 32'hffff_ffff;
		end
		else begin
			wr_1l	<= ~xtwr;
			tbe_1l	<= (xtbe[3:0] == 4'b0000);
			cs_1l	<= ~xtcs;
			ta_1l	<= ta;
			td_1l	<= td;
		end
	end

/***************************************************************************/
/* pipeline 2'nd                                                           */
/***************************************************************************/
	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			wr_2l	<= 1'b0;
			cs_2l	<= 1'b0;
			ta_2l	<= 3'b0;
			td_2l	<= 32'hffff_ffff;
		end
		else begin
			wr_2l	<= wr_1l & tbe_1l;
			cs_2l	<= cs_1l;
			ta_2l	<= ta_1l;
			td_2l	<= td_1l;
		end
	end

/***************************************************************************/
/* pipeline 3'rd                                                           */
/***************************************************************************/
	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			wr_3l	<= 1'b0;
			cs_3l	<= 1'b0;
			ta_3l	<= 3'b0;
			td_3l	<= 32'hffff_ffff;
		end
		else begin
			wr_3l	<= wr_2l;
			cs_3l	<= cs_2l;
			ta_3l	<= ta_2l;
			td_3l	<= td_2l;
		end
	end

/***************************************************************************/
/* pipeline 4'th                                                           */
/***************************************************************************/
	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			sync_wr	  <= 1'b0;
			sync_data <= 36'hf_ffff_ffff;
		end
		else begin
			sync_wr	<= ~wr_msk & cs_3l & ~wr_2l & wr_3l;
			sync_data[35:4]	<= td_3l[31:0];
			sync_data[3:1]	<= ta_3l[4:2];
			sync_data[0]	<= 1'b0;
		end
	end

/***************************************************************************/
/* fifo full detector                                                      */
/***************************************************************************/

	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			empty_1l  <= 1'b1;
			empty_2l  <= 1'b1;
			full_sync <= 1'b0;
		end
		else begin
			empty_1l  <= empty;
			empty_2l  <= empty_1l;
			full_sync <= ~empty_2l & full;
		end
	end


	always @(posedge busclk or posedge fifo_rst) begin
		if (fifo_rst) begin
			full_lat  <= 1'b0;
		end
		else begin
			if (full_sync == 1'b1) begin
				full_lat  <= 1'b1;
			end
			else begin
				full_lat  <= full_lat;
			end
		end
	end

/***************************************************************************/
/* header detector                                                         */
/***************************************************************************/

	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			header_det <= 1'b0;
		end
		else begin
			if (ta_1l[2] == 1'b0) begin
				header_det <= wr_1l;
			end
			else begin
				header_det <= 1'b0;
			end
		end
	end


/***************************************************************************/
/* fifo write mask                                                         */
/***************************************************************************/

	always @(posedge busclk or posedge reset) begin
		if (reset) begin
			wr_msk		<= 1'b0;
			msk_count	<= 2'b11;
			write_msk_st	<= WAIT_FULL;
		end
		else begin
			case (write_msk_st)
			WAIT_FULL : begin
				msk_count	<= 2'b11;
				if (full_lat == 1'b1) begin
					wr_msk	<= 1'b1;
					write_msk_st	<= WAIT_FULL_NEG;
				end
				else begin
					wr_msk	<= 1'b0;
					write_msk_st	<= WAIT_FULL;
				end
			end
			WAIT_FULL_NEG : begin
				wr_msk	<= 1'b1;
				msk_count	<= 2'b11;
				if (full_lat == 1'b0) begin
					write_msk_st	<= WRITE_MSK;
				end
				else begin
					write_msk_st	<= WAIT_FULL_NEG;
				end
			end
			WRITE_MSK : begin
				wr_msk	<= 1'b1;
				msk_count	<= msk_count - 1;
				if (msk_count == 2'b00) begin
					write_msk_st	<= WAIT_HEADER;
				end
				else begin
					write_msk_st	<= WRITE_MSK;
				end
			end
			WAIT_HEADER : begin
				msk_count	<= 2'b11;
				if (header_det == 1'b1) begin
					wr_msk	<= 1'b1;
					write_msk_st	<= WAIT_FULL;
				end
				else begin
					wr_msk	<= 1'b1;
					write_msk_st	<= WAIT_HEADER;
				end
			end
			endcase
		end
	end

endmodule
